@@ -1,14 +1,16 @@
+import copy
 import json
 import requests
 import warnings
 
-from sigopt.endpoint import ApiEndpoint
-from sigopt.exception import ApiException
-from sigopt.objects import ApiObject
-from sigopt.resource import ApiResource
-from sigopt.response import (
+from .endpoint import ApiEndpoint
+from .exception import ApiException
+from .objects import ApiObject
+from .resource import ApiResource
+from .response import (
   ExperimentResponse, ClientResponse,
   ExperimentsBestObservationResponse, ExperimentsSuggestResponse, ExperimentsWorkersResponse,
+  ExperimentsAllocateResponse, ExperimentsCreateCohortResponse, ExperimentsUpdateCohortResponse,
   ClientsExperimentsResponse,
 )
 
@@ -18,6 +20,11 @@ class Connection(object):
     if client_token is None and user_token is None:
       raise ValueError('Must provide either user_token or client_token (or both)')
 
+    self.default_params = {
+      'user_token': user_token,
+      'client_token': client_token,
+      'worker_id': worker_id,
+    }
     self.client_token = client_token
     self.user_token = user_token
     self.worker_id = worker_id
@@ -27,13 +34,16 @@ class Connection(object):
       'experiments',
       response_cls=ExperimentResponse,
       endpoints=[
+        ApiEndpoint('allocate', ExperimentsAllocateResponse, 'GET'),
         ApiEndpoint('bestobservation', ExperimentsBestObservationResponse, 'GET'),
+        ApiEndpoint('createcohort', ExperimentsCreateCohortResponse, 'POST'),
         ApiEndpoint('delete', None, 'POST'),
         ApiEndpoint('releaseworker', None, 'POST'),
         ApiEndpoint('report', None, 'POST'),
         ApiEndpoint('reportmulti', None, 'POST'),
         ApiEndpoint('suggest', ExperimentsSuggestResponse, 'POST'),
         ApiEndpoint('update', ExperimentResponse, 'POST'),
+        ApiEndpoint('updatecohort', ExperimentsUpdateCohortResponse, 'POST'),
         ApiEndpoint('workers', ExperimentsWorkersResponse, 'GET'),
       ],
     )
@@ -98,8 +108,8 @@ class Connection(object):
       response = response_json.get('response')
       return response
     else:
-      error_message = response_json.get('error', {}).get('message', None)
-      raise ApiException(error_message, response.status_code)
+      error_json = response_json.get('error', {})
+      raise ApiException(error_json, response.status_code)
 
   def _ensure_client_token(self):
     if self.client_token is None:
@@ -110,19 +120,20 @@ class Connection(object):
       raise ValueError('user_token is required for this call')
 
   def _request_params(self, params):
-    ret = {
-      'user_token': self.user_token,
-      'client_token': self.client_token,
-      'worker_id': self.worker_id,
-    }
-    ret.update(params or {})
+    req_params = copy.copy(self.default_params)
+    req_params.update(params or {})
 
     def serialize(value):
       if isinstance(value, dict):
         return json.dumps(value)
       return str(value)
 
-    return dict(((key, serialize(self._to_api_value(value))) for key, value in ret.iteritems()))
+    return dict((
+      (key, serialize(self._to_api_value(value)))
+      for key, value
+      in req_params.iteritems()
+      if value is not None
+    ))
 
   def _to_api_value(self, obj):
     if isinstance(obj, ApiObject):
